// =====================================================================================================================
//  fennec, a free and open source game engine
//  Copyright © 2025  Medusa Slockbower
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.
// =====================================================================================================================

///
/// \file sequences.h
/// \brief \ref fennec_lang_sequences
///
///
/// \details
/// \author Medusa Slockbower
///
/// \copyright Copyright © 2025 Medusa Slockbower ([GPLv3](https://www.gnu.org/licenses/gpl-3.0.en.html))
///
///

#ifndef FENNEC_LANG_SEQUENCES_H
#define FENNEC_LANG_SEQUENCES_H

///
/// \page fennec_lang_sequences Sequences
///
/// \brief This header is part of the metaprogramming library. It defines structures for sequences of values, used during compile time.
///
/// \code #include <fennec/lang/sequences.h> \endcode
///
/// <table width="100%" class="fieldtable" id="table_fennec_lang_constants">
/// <tr><th style="vertical-align: top">Syntax
///     <th style="vertical-align: top">Description
///
/// <tr><td width="50%" style="vertical-align: top"> <br>
///         \ref fennec::sequence "sequence<ValueT, Values...>"<br>
///     <td width="50%" style="vertical-align: top">
///         \copydetails fennec::sequence
///
/// <tr><td width="50%" style="vertical-align: top"> <br>
///         \ref fennec::integer_sequence "integer_sequence<IntT, Values...>"<br>
///         \ref fennec::make_integer_sequence "typename make_integer_sequence<IntT, N>::type"<br>
///         \ref fennec::make_integer_sequence_t "make_integer_sequence_t<IntT, N>"
///     <td width="50%" style="vertical-align: top">
///         \copydetails fennec::integer_sequence
///
/// <tr><td width="50%" style="vertical-align: top"> <br>
///         \ref fennec::index_sequence "index_sequence<Indices...>"<br>
///         \ref fennec::make_index_sequence "typename make_index_sequence<N>::type"<br>
///         \ref fennec::make_index_sequence_t "make_index_sequence_t<N>"
///     <td width="50%" style="vertical-align: top">
///         \copydetails fennec::index_sequence
///
/// <tr><td width="50%" style="vertical-align: top"> <br>
///         \ref fennec::concat_sequence "typename concat_sequence<SequenceT0, SequenceT1>::type"<br>
///         \ref fennec::concat_sequence_t "concat_sequence_t<SequenceT0, SequenceT1>"<br>
///     <td width="50%" style="vertical-align: top">
///         \copydetails fennec::concat_sequence
///
/// </table>
///

#include <fennec/lang/type_traits.h>

namespace fennec
{

// fennec::sequence ====================================================================================================

///
/// \brief metaprogramming sequence
///
/// \details Stores a sequence of values of type `ValueT` as a template pack.
///          You can access the parameter pack in another template function, i.e.
///          \code{cpp}
///          template<typename TypeT, TypeT...Values>
///          constexpr TypeT summation(sequence<TypeT, Values...>)
///          {
///              return (Values + ...);
///          }
///          \endcode
/// \tparam ValueT type of the values
/// \tparam Values sequence values
template<typename ValueT, ValueT...Values> struct sequence
{
	/// \brief type of the sequence
	using value_type = ValueT;

	/// \brief self-referential type
	using type = sequence;

	///
	/// \brief returns the number of elements
	///
	/// \return number of elements in the array
	inline static constexpr size_t size() noexcept { return sizeof...(Values); }
};



// fennec::integer_sequence ============================================================================================

///
/// \brief metaprogramming integral sequence
///
/// \details A `fennec::sequence` specialized integer types.
/// \tparam IntT type of the values, must satisfy ```fennec::is_integral<T>```
/// \tparam Values sequence values
template<typename IntT, IntT...Values> requires(is_integral_v<IntT>)
struct integer_sequence : sequence<IntT, Values...>
{
	/// \brief type of the sequence
	using value_type = IntT;

	/// \brief self-referential type
	using type = integer_sequence;

	///
	/// \brief returns the number of elements
	///
	/// \return number of elements in the array
	inline static constexpr size_t size() noexcept { return sizeof...(Values); }
};


///
/// \brief generate a fennec::integer_sequence \f$\left[\,0\,\ldots\,N\,\right)\f$
///
/// \details
/// \tparam IntT type of the values, must satisfy ```fennec::is_integral<T>```
/// \tparam N size of the sequence to generate
template<typename IntT, size_t N> struct make_integer_sequence;

///
/// \brief shorthand for ```typename make_integer_sequence<T, N>::type```
template<typename IntT, size_t N> using  make_integer_sequence_t = typename make_integer_sequence<IntT, N>::type;



// fennec::index_sequence ==============================================================================================

///
/// \brief metaprogramming integral sequence
///
/// \details A `fennec::integer_sequence` specialized for sequences of `size_t` indices.
/// \tparam Indices sequence values
template<size_t...Indices> struct index_sequence : integer_sequence<size_t, Indices...>
{
	/// \brief type of the sequence
	using value_type = size_t;

	/// \brief self-referential type
	using type = index_sequence;

	///
	/// \brief returns the number of elements
	///
	/// \return number of elements in the array
	inline static constexpr size_t size() noexcept { return sizeof...(Indices); }
};


///
/// \brief generate a fennec::index_sequence \f$\left[\,0\,\ldots\,N\,\right)\f$
///
/// \details
/// \tparam T type of the values, must satisfy ```fennec::is_integral<T>```
/// \tparam N size of the sequence to generate
template<size_t N> struct make_index_sequence;

///
/// \brief shorthand for ```typename make_index_sequence<N>::type```
template<size_t N> using  make_index_sequence_t = typename make_index_sequence<N>::type;



// fennec::concat_sequence =============================================================================================

///
/// \brief concatenate two sequences
///
/// \details A tool for concatenating two `fennec::sequence` types.
/// \tparam SequenceT0 lhs
/// \tparam SequenceT1 rhs
template<typename SequenceT0, typename SequenceT1> struct concat_sequence;

///
/// \brief shorthand for ```typename concat_sequence<SequenceT0, SequenceT1>::type```
template<typename SequenceT0, typename SequenceT1> using concat_sequence_t
	= typename concat_sequence<SequenceT0, SequenceT1>::type;



// Internal ============================================================================================================

// Implementation for Generating an \ref integer_sequence
template<typename T, size_t N> struct make_integer_sequence : concat_sequence_t<make_integer_sequence_t<T, N / 2>, make_integer_sequence_t<T, N - N / 2>>{};

// Base Case of \f$N=0\f$
template<typename T> struct make_integer_sequence<T, 0> : integer_sequence<T> {};

// Base Case of \f$N=1\f$
template<typename T> struct make_integer_sequence<T, 1> : integer_sequence<T, 0>{};


// Implementation for Generating an \ref integer_sequence
template<size_t N> struct make_index_sequence : concat_sequence_t<make_index_sequence_t<N / 2>, make_index_sequence_t<N - N / 2>>{};

// Base Case of \f$N=0\f$
template<> struct make_index_sequence<0> : index_sequence<> {};

// Base Case of \f$N=1\f$
template<> struct make_index_sequence<1> : index_sequence<0>{};


// Specialization for integer sequences
template<typename T, T...SequenceV0, T...SequenceV1>
struct concat_sequence<integer_sequence<T, SequenceV0...>, integer_sequence<T, SequenceV1...>>
	: integer_sequence<T, SequenceV0..., (sizeof...(SequenceV0) + SequenceV1)...>{};

// Specialization for index sequences
template<size_t...SequenceV0, size_t...SequenceV1>
struct concat_sequence<index_sequence<SequenceV0...>, index_sequence<SequenceV1...>>
	: index_sequence<SequenceV0..., (sizeof...(SequenceV0) + SequenceV1)...>{};



}

#endif // FENNEC_LANG_SEQUENCES_H
