// =====================================================================================================================
//  fennec, a free and open source game engine
//  Copyright © 2025  Medusa Slockbower
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <https://www.gnu.org/licenses/>.
// =====================================================================================================================

///
/// \file utility.h
/// \brief \ref fennec_lang_utility
///
///
/// \details
/// \author Medusa Slockbower
///
/// \copyright Copyright © 2025 Medusa Slockbower ([GPLv3](https://www.gnu.org/licenses/gpl-3.0.en.html))
///
///

#ifndef FENNEC_LANG_UTILITY_H
#define FENNEC_LANG_UTILITY_H

///
/// \page fennec_lang_utility Utility
///
/// \brief This header contains common utility functions related to the C++ language.
///
/// \code #include <fennec/lang/utility.h> \endcode
///
/// <table width="100%" class="fieldtable" id="table_fennec_lang_constants">
/// <tr><th style="vertical-align: top">Syntax
///     <th style="vertical-align: top">Description
///
/// <tr><td width="50%" style="vertical-align: top"> <br>
///         \ref fennec::forward "T&& forward(x)"<br>
///     <td width="50%" style="vertical-align: top">
///         \copydetails fennec::forward
///
/// <tr><td width="50%" style="vertical-align: top"> <br>
///         \ref fennec::move "T&& move(x)"<br>
///     <td width="50%" style="vertical-align: top">
///         \copydetails fennec::move
///
/// <tr><td width="50%" style="vertical-align: top"> <br>
///         \ref fennec::copy "const T& copy(x)"<br>
///     <td width="50%" style="vertical-align: top">
///         \copydetails fennec::copy
///
/// </table>
///

#include <fennec/lang/type_transforms.h>

namespace fennec
{

///
/// \brief forwards reference types to extend their lifetime
///
/// \details forwards reference types to extend their lifetime
/// \tparam T base type of the object
/// \param x reference to the object
/// \returns
template<typename T> constexpr T&& forward(remove_reference_t<T>& x) noexcept { return x; }

// specialization for T&&
template<typename T> constexpr T&& forward(remove_reference_t<T>&& x) noexcept { return x; }


///
/// \brief produces an x-value type to indicate \p x may be "moved"
///
/// \details produces an x-value type to indicate \p x may be "moved"
/// \tparam T base type of the object
/// \param x object to be moved
/// \returns `static_cast<remove_reference_t<T>&&>(x)`
template<typename T> constexpr remove_reference_t<T>&& move(T&& x) noexcept { return static_cast<remove_reference_t<T>&&>(x); }


///
/// \brief produces an r-value type to indicate \p x may be "copied"
///
/// \details produces an r-value type to indicate \p x may be "copied"
/// \tparam T base type of the object
/// \param x object to be copied
/// \returns const r-value
template<typename T> constexpr const remove_reference_t<T>& copy(T&& x) noexcept { return x; }

}

#endif // FENNEC_LANG_UTILITY_H
